//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_MESSAGES_H
#define PROTOCOL_MESSAGES_H

#include <cstdint>
#include <vector>
#include <boost/system/error_code.hpp>
#include <boost/asio/write.hpp>
#include <boost/asio/read.hpp>
#include "Utils.h"
#include "Networking.h"

namespace networking
{
namespace message
{

template<typename Message, typename SyncWriteStream>
void send(SyncWriteStream & socket, const Message & message)
{
    boost::asio::write(socket, boost::asio::buffer(message.data()));
};

template<typename Message, typename SyncWriteStream>
void send(Networking & net,
          SyncWriteStream & socket,
          const Message & message,
          const boost::posix_time::time_duration & timeout)
{
    if (timeout <= boost::posix_time::seconds(0))
        throw std::runtime_error("message::send() given timeout <= 0");

    networking::utils::write(net, socket, boost::asio::buffer(message.data()), timeout);
};

template<typename Message, typename SyncReadStream>
Message receive(SyncReadStream & socket)
{
    auto buf = Message::emptyBuffer();
    boost::asio::read(socket, boost::asio::buffer(buf));
    return Message(buf);
};

template<typename Message, typename SyncReadStream>
Message receive(Networking & net,
                SyncReadStream & socket,
                const boost::posix_time::time_duration & timeout)
{
    if (timeout <= boost::posix_time::seconds(0))
        throw std::runtime_error("message::receive() given timeout <= 0");

    auto buf = Message::emptyBuffer();
    networking::utils::read(net, socket, boost::asio::buffer(buf), timeout);
    return Message(buf);
};

template<typename Message, typename DatagramSocket, typename Endpoint>
void sendDatagram(DatagramSocket & socket, const Message & message, const Endpoint & endpoint)
{
    auto buf = Message::emptyBuffer();
    socket.send_to(buf, endpoint);
}

template<typename Message>
class DatagramSender : private boost::noncopyable
{
public:
    using Socket = boost::asio::ip::udp::socket;
    using Endpoint = boost::asio::ip::udp::endpoint;

    DatagramSender(Networking & net)
        : net(net)
          , socket(net.getIoService())
    {}

    void send(const Message & message, const std::string & host, uint16_t port)
    {
        Endpoint endpoint{boost::asio::ip::address::from_string(host), port};
        sendDatagram(socket, message, endpoint);
    }

private:
    Networking & net;
    Socket socket;
};

template<typename Message, typename DatagramSocket, typename Endpoint, typename ReceiveHandler>
void asyncDatagramReceive(DatagramSocket & socket,
                          Endpoint & senderEndpoint,
                          ReceiveHandler receiveHandler,
                          const boost::posix_time::time_duration & timeout)
{
    auto buf = std::make_shared<Message>(Message::emptyBuffer());
    networking::utils::asyncReceiveFrom(socket, boost::asio::buffer(*buf), senderEndpoint,
                                        [buf, senderEndpoint, receiveHandler](
                                            const boost::system::error_code & receiveError,
                                            std::size_t bytesTransferred)
                                        {
                                            Message msg(*buf);
                                            boost::system::error_code error = receiveError;
                                            if (receiveError && bytesTransferred < buf->size())
                                                error = boost::asio::error::message_size;
                                            receiveHandler(senderEndpoint, msg, error);
                                        }, timeout);

};

template<typename Message>
class AsyncDatagramReceiver
{
public:
    using Endpoint = boost::asio::ip::udp::endpoint;
    using Socket = boost::asio::ip::udp::socket;
    using ReceiveHandler = std::function<
        void(const Endpoint & senderEndpoint, const Message & message, boost::system::error_code error)>;

    AsyncDatagramReceiver(Networking & net)
        : net(net)
          , socket(net.getIoService())
    {}

    AsyncDatagramReceiver(const AsyncDatagramReceiver &) = delete;

    AsyncDatagramReceiver & operator=(const AsyncDatagramReceiver &) = delete;

    AsyncDatagramReceiver(AsyncDatagramReceiver &&) = delete;

    AsyncDatagramReceiver & operator=(AsyncDatagramReceiver &&) = delete;

    void receive(const std::string & host,
                 std::uint16_t port,
                 ReceiveHandler receiveHandler,
                 const boost::posix_time::time_duration & timeout)
    {
        Endpoint endpoint{boost::asio::ip::address::from_string(host), port};
        socket.bind(endpoint);
        asyncDatagramReceive(socket, senderEndpoint, receiveHandler, timeout);
    }

private:
    Networking & net;
    Endpoint senderEndpoint;
    Socket socket;
};

}
}


#endif //PROTOCOL_MESSAGES_H
