//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_ProtocolNETWORKSERVICESERVER_H
#define PROTOCOL_ProtocolNETWORKSERVICESERVER_H

#include "Message.h"
#include "Networking.h"

namespace networking
{
namespace service
{

template<typename Service>
class Server : public std::enable_shared_from_this<Server<Service>>
{
private:
    struct PrivateTag
    {
    };

public:
    using RequestMessage = typename Service::RequestMessage;
    using ResponseMessage = typename Service::ResponseMessage;

    using Ptr = std::shared_ptr<Server<Service>>;

    using Endpoint = typename boost::asio::ip::tcp::endpoint;
    using Socket = typename boost::asio::ip::tcp::socket;
    using Acceptor = typename boost::asio::ip::tcp::acceptor;

    using OnRequestReceivedCallback =
    std::function<ResponseMessage(const Endpoint & clientEndpoint, const RequestMessage & requestMessage)>;

    using OnNetworkErrorCallback = std::function<void(const boost::system::system_error & error)>;

    static Ptr create(Networking & net)
    {
        return std::make_shared<Server<Service>>(PrivateTag{}, net);
    }

    // Should not be used outside.
    Server(PrivateTag, Networking & net)
        : net(net)
          , acceptor(net.getIoService())
    {}

    ~Server()
    {
        stop();
    }

    Server(const Server &) = delete;

    Server & operator=(const Server &) = delete;

    Server(Server &&) = delete;

    Server & operator=(Server &&) = delete;

    void advertiseService(const Endpoint & bindEndpoint,
                          OnRequestReceivedCallback onRequestReceivedCallback,
                          OnNetworkErrorCallback onNetworkErrorCallback = defaultOnNetworkErrorCallback)
    {
        if (running)
            stop();

        running = true;

        this->onRequestReceivedCallback = onRequestReceivedCallback;
        this->onNetworkErrorCallback = onNetworkErrorCallback;

        acceptor = Acceptor(net.getIoService(), bindEndpoint);
        accept();
    }

    void advertiseService(const std::string & host, uint16_t port,
                          OnRequestReceivedCallback onRequestReceivedCallback,
                          OnNetworkErrorCallback onNetworkErrorCallback = defaultOnNetworkErrorCallback)
    {
        advertiseService(Endpoint(
            boost::asio::ip::address::from_string(host), port),
                         onRequestReceivedCallback, onNetworkErrorCallback);
    }

    void stop()
    {
        running = false;
        boost::system::error_code ignoredError;
        acceptor.close(ignoredError);
    }

private:
    Networking & net;
    OnRequestReceivedCallback onRequestReceivedCallback;
    OnNetworkErrorCallback onNetworkErrorCallback;
    Acceptor acceptor;
    std::atomic<bool> running{false};

    void accept()
    {
        auto self = this->shared_from_this();
        auto socket = std::make_shared<Socket>(net.getIoService());
        acceptor.async_accept(*socket, [self, socket](boost::system::error_code acceptError)
        {
            if (!self->running)
                return;

            if (acceptError)
            {
                self->onNetworkErrorCallback(acceptError);
                return;
            }

            try
            {
                auto request = message::receive<RequestMessage>(self->net, *socket, boost::posix_time::seconds(3));
                auto response = self->onRequestReceivedCallback(socket->remote_endpoint(), request);
                message::send(self->net, *socket, response, boost::posix_time::seconds(3));
            }
            catch (const boost::system::system_error & error)
            {
                self->onNetworkErrorCallback(error);
            }

            // The next accept event will be put on the event queue.
            self->accept();
        });
    }

    //
    static void defaultOnNetworkErrorCallback(const boost::system::system_error & error)
    {}
};

}
}

#endif //PROTOCOL_ProtocolNETWORKSERVICESERVER_H
