//
// Created by philipp on 02.12.17.
//

#ifndef PROTOCOL_NETWORKUTILS_H
#define PROTOCOL_NETWORKUTILS_H

#include <boost/asio/ip/tcp.hpp>
#include <boost/asio.hpp>
#include "Networking.h"

namespace networking
{
namespace utils
{

// Little wrapper for adding an enabled flag to the deadline_timer.
class DeadlineTimer : public std::enable_shared_from_this<DeadlineTimer>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<DeadlineTimer>;
    using TimeoutCallback = std::function<void()>;

    DeadlineTimer(PrivateTag, boost::asio::io_service & ioService)
        : timer(ioService)
    {}

    static Ptr create(boost::asio::io_service & ioService)
    {
        return std::make_shared<DeadlineTimer>(PrivateTag{}, ioService);
    }

    DeadlineTimer(const DeadlineTimer &) = delete;

    DeadlineTimer & operator=(const DeadlineTimer &) = delete;

    DeadlineTimer(DeadlineTimer &&) = delete;

    DeadlineTimer & operator=(DeadlineTimer &&) = delete;

    void asyncWait(const boost::posix_time::time_duration & timeout, TimeoutCallback callback)
    {
        auto self = shared_from_this();
        timer.expires_from_now(timeout);
        timer.async_wait([self, callback](const boost::system::error_code & error)
                         {
                             if (error || !self->enabled)
                                 return;

                             callback();
                         });
    }

    void disable()
    {
        enabled = false;
        timer.cancel();
    }

private:
    boost::asio::deadline_timer timer;
    std::atomic<bool> enabled{true};
};

template<typename AsyncOperation, typename Socket, typename... Args>
void timedAsyncSocketOperation(Networking & net,
                               AsyncOperation asyncOperation,
                               Socket & socket,
                               const boost::posix_time::time_duration & timeout,
                               Args && ... args)
{
    auto & ioService = net.getIoService();

    auto deadline = DeadlineTimer::create(ioService);
    deadline->asyncWait(timeout,
                        [&socket]()
                        {
                            // Timeout expired: close the socket.
                            boost::system::error_code ignoredError;
                            socket.close(ignoredError);
                        });

    // A 'would_block' socketError is guaranteed to never occur on an asynchronous operation.
    boost::system::error_code socketError = boost::asio::error::would_block;

    // Run asynchronous connect.
    asyncOperation(socket, std::forward<Args>(args)...,
                   [&socketError, deadline](const boost::system::error_code & error, auto ...)
                   {
                       deadline->disable();
                       // update socketError variable
                       socketError = error;
                   });

    // Wait until "something happens" with the socket.
    net.waitUntil([&socketError]()
                  { return socketError != boost::asio::error::would_block; });

    // Determine whether a connection was successfully established.
    // Even though our deadline handler might have run to close the socket, the connect operation
    // might have notionally succeeded!
    if (socketError || !socket.is_open())
    {
        throw boost::system::system_error(
            socketError ? socketError : boost::asio::error::operation_aborted);
    }
}

template<typename Socket, typename EndpointIterator>
void connect(Networking & net,
             Socket & socket,
             const EndpointIterator & endpointIterator,
             const boost::posix_time::time_duration & timeout)
{
    // This lambda is used because we cannot pass boost::asio::async_connect directly as a function
    // since it is a template.
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, socket, timeout, endpointIterator);
}

template<typename SyncWriteStream, typename ConstBufferSequence>
void write(Networking & net,
           SyncWriteStream & stream,
           const ConstBufferSequence & buffer,
           const boost::posix_time::time_duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, stream, timeout, buffer);
};

template<typename SyncReadStream, typename MutableBufferSequence>
void read(Networking & net,
          SyncReadStream & stream,
          const MutableBufferSequence & buffer,
          const boost::posix_time::time_duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, stream, timeout, buffer);
};

template<
    typename DatagramSocket,
    typename MutableBuffers,
    typename Endpoint,
    typename ReceiveHandler>
void asyncReceiveFrom(DatagramSocket & socket,
                      MutableBuffers & buffer,
                      Endpoint & endpoint,
                      ReceiveHandler receiveHandler,
                      const boost::posix_time::time_duration & timeout)
{
    auto & ioService = socket.get_io_service();

    auto deadline = DeadlineTimer::create(ioService);
    deadline->asyncWait(timeout,
                        [&socket]()
                        {
                            // Timeout expired: close the socket.
                            boost::system::error_code ignoredError;
                            socket.close(ignoredError);
                        });

    socket.async_receive_from(buffer, endpoint,
                              [deadline, receiveHandler](
                                  const boost::system::error_code & receiveError,
                                  std::size_t bytesTransferred)
                              {
                                  deadline->disable();
                                  receiveHandler(receiveError, bytesTransferred);
                              }
    );
}

}
}

#endif //PROTOCOL_NETWORKUTILS_H
