//
// Created by philipp on 30.11.17.
//

#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/FollowerVehicle.h"
#include "../include/Utils.h"

using namespace protocol;

void FollowerVehicle::doOnCreatePlatoon()
{
    log("follower[", std::to_string(vehicleId), "] starts creating platoon\n");

    using namespace std::placeholders;
    platoonCreateServiceServer->advertiseService(
        "127.0.0.1",
        vehicleId + 10000,  // TODO: endpoint information!
        std::bind(&FollowerVehicle::makePlatoonCreateResponse, this, _1, _2));

    scanAvailableVehicles();
    vehicleEndpointIter = availableVehicleEndpoints.begin();

    state = State::CREATING_PLATOON;
}

void FollowerVehicle::doCreatePlatoon()
{
    {
        std::lock_guard<std::mutex> lock(leaderMutex);
        if (leaderFound)
        {
            state = State::ON_RUNNING_PLATOON;
            return;
        }
    }

    auto vehicleEndpoint = *vehicleEndpointIter;
    makePlatoonCreateRequest(vehicleEndpoint);
    utils::cycle(vehicleEndpointIter, availableVehicleEndpoints);
}

void FollowerVehicle::doOnRunningPlatoon()
{
    log("follower[", std::to_string(vehicleId), "] starts running a platoon\n");
    platoonCreateServiceServer->stop();
    state = State::RUNNING_PLATOON;
}

void FollowerVehicle::doRunningPlatoon()
{
    log("follower[", std::to_string(vehicleId), "] has leader[", std::to_string(leader.getVehicleId()),
                 "] and platoon id: ", std::to_string(platoonId), "\n");
    // TODO: implement running behavior
    state = State::ON_LEAVING_PLATOON;
}

void FollowerVehicle::doOnLeavingPlatoon()
{
    platoonCreateServiceServer->stop();
    state = State::LEAVING_PLATOON;
}

void FollowerVehicle::doLeavingPlatoon()
{
    state = State::IDLE;
}

void FollowerVehicle::makePlatoonCreateRequest(VehicleEndpoint vehicleEndpoint)
{
    try
    {
        // We'll send a request indicating that we want to be the follower.
        auto request = PlatoonCreateNetworkMessage::followerRequest(vehicleId);
        auto response = platoonCreateServiceClient.call(vehicleEndpoint.getHost(),
                                    vehicleEndpoint.getTcpPort(),
                                    request,
                                    boost::posix_time::seconds(3));
        // Verify that the vehicle-id is correct.
        // If our request was accepted, we found our follower.
        if (response.getVehicleId() == vehicleEndpoint.getVehicleId() &&
            response.getMessageType() == messageTypes::ACCEPT_RESPONSE)
        {
            std::lock_guard<std::mutex> lock(leaderMutex);
            leader = vehicleEndpoint;
            leaderFound = true;
            platoonId = response.getPlatoonId();
        }
    }
    catch (const std::exception & e)
    {
        // This maybe a timeout or unable to connect to the other vehicle.
        // We just keep retrying in the next runs, so we don't need to handle this any further.
    }
}

PlatoonCreateNetworkMessage FollowerVehicle::makePlatoonCreateResponse(
    const tcp::endpoint & clientEndpoint, const PlatoonCreateNetworkMessage & request)
{
    // Only consider this message if it is a leader request, meaning that a leader wants us to join its platoon.
    if (request.getMessageType() != messageTypes::LEADER_REQUEST)
        return PlatoonCreateNetworkMessage::rejectResponse(vehicleId);

    std::lock_guard<std::mutex> lock(leaderMutex);
    if (leaderFound)
        return PlatoonCreateNetworkMessage::rejectResponse(vehicleId);

    auto otherVehicleId = request.getVehicleId();
    // TODO: change vehicle endpoint information!
    leader = VehicleEndpoint{
        otherVehicleId,
        clientEndpoint.address().to_string(),
        otherVehicleId + 10000,
        otherVehicleId + 10010
    };
    leaderFound = true;
    platoonId = request.getPlatoonId();

    return PlatoonCreateNetworkMessage::acceptResponse(vehicleId, platoonId);
}
