//
// Created by philipp on 30.11.17.
//

#include <iostream>
#include "../include/LeaderVehicle.h"
#include "../include/Utils.h"

using namespace protocol;

void LeaderVehicle::doOnCreatePlatoon()
{
    log("leader[", std::to_string(vehicleId), "] starts creating platoon\n");
    // Generate a platoon id.
    platoonId = rand();

    using namespace std::placeholders;
    platoonCreateServiceServer->advertiseService(
        "127.0.0.1",
        vehicleId + 10000,  // TODO: endpoint information!
        std::bind(&LeaderVehicle::makePlatoonCreateResponse, this, _1, _2));

    scanAvailableVehicles();
    vehicleEndpointIter = availableVehicleEndpoints.begin();

    state = State::CREATING_PLATOON;
}

void LeaderVehicle::doCreatePlatoon()
{
    {
        std::lock_guard<std::mutex> lock(followersMutex);
        if (!followers.empty())
        {
            state = State::ON_RUNNING_PLATOON;
            return;
        }
    }

    auto vehicleEndpoint = *vehicleEndpointIter;
    makePlatoonCreateRequest(vehicleEndpoint);
    utils::cycle(vehicleEndpointIter, availableVehicleEndpoints);
}

void LeaderVehicle::doOnRunningPlatoon()
{
    log("leader[", std::to_string(vehicleId), "] starts running a platoon\n");

    state = State::RUNNING_PLATOON;
}

void LeaderVehicle::doRunningPlatoon()
{
    log("leader[", std::to_string(vehicleId), "] has platoon id: ", std::to_string(platoonId), "\n");
    {
        std::lock_guard<std::mutex> lock(followersMutex);
        for (auto & pair : followers)
            log("leader[", std::to_string(vehicleId), "] has follower: ", std::to_string(pair.first), "\n");
    }
    // TODO: implement running behavior
}

void LeaderVehicle::doOnLeavingPlatoon()
{
    state = State::LEAVING_PLATOON;
}

void LeaderVehicle::doLeavingPlatoon()
{
    state = State::IDLE;
}

void LeaderVehicle::makePlatoonCreateRequest(VehicleEndpoint vehicleEndpoint)
{
    try
    {
        // We'll send a request indicating that we want to be the leader.
        auto request = PlatoonCreateNetworkMessage::leaderRequest(vehicleId, platoonId);
        auto response = platoonCreateServiceClient.call(vehicleEndpoint.getHost(),
                                                        vehicleEndpoint.getTcpPort(),
                                                        request,
                                                        boost::posix_time::seconds(1));
        // Verify response.
        if (response.getVehicleId() == vehicleEndpoint.getVehicleId() &&
            response.getMessageType() == messageTypes::ACCEPT_RESPONSE &&
            response.getPlatoonId() == platoonId)
        {
            std::lock_guard<std::mutex> lock(followersMutex);
            followers[response.getVehicleId()] = vehicleEndpoint;
        }
    }
    catch (const std::exception & e)
    {
        // This maybe a timeout or unable to connect to the other vehicle.
        // We just keep retrying in the next runs, so we don't need to handle this any further.
    }
}

PlatoonCreateNetworkMessage LeaderVehicle::makePlatoonCreateResponse(const tcp::endpoint & clientEndpoint,
                                                                     const PlatoonCreateNetworkMessage & request)
{
    // Only consider this message if it is a follower request, meaning that a follower wants us to be its leader.
    if (request.getMessageType() != messageTypes::FOLLOWER_REQUEST)
        return PlatoonCreateNetworkMessage::rejectResponse(vehicleId);

    std::lock_guard<std::mutex> lock(followersMutex);
    auto otherVehicleId = request.getVehicleId();
    // TODO: change vehicle endpoint information!
    followers[otherVehicleId] = VehicleEndpoint{
        otherVehicleId,
        clientEndpoint.address().to_string(),
        otherVehicleId + 10000,
        otherVehicleId + 10010};

    return PlatoonCreateNetworkMessage::acceptResponse(vehicleId, platoonId);
}

std::size_t LeaderVehicle::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(followersMutex);
    return followers.size();
}

void LeaderVehicle::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    std::lock_guard<std::mutex> lock(platoonConfigMutex);
    this->platoonConfig = platoonConfig;
}


